@@ -48,6 +48,7 @@
 #include <rpc/rpcsec_gss.h>
 #include <rpc/clnt.h>
 #include <netinet/in.h>
+#include <reentrant.h>
 
 #include "debug.h"
 
@@ -134,6 +135,8 @@ char *p;
 	fprintf(stderr, "     cred: %p\n", ptr->cred);
 }
 
+extern pthread_mutex_t		 auth_ref_lock;
+
 struct rpc_gss_data {
 	bool_t			 established;	/* context established */
 	gss_buffer_desc		 gc_wire_verf;	/* save GSS_S_COMPLETE NULL RPC verfier
@@ -144,12 +147,37 @@ struct rpc_gss_data {
 	gss_ctx_id_t		 ctx;		/* context id */
 	struct rpc_gss_cred	 gc;		/* client credentials */
 	u_int			 win;		/* sequence window */
+	int			 time_req;	/* init_sec_context time_req */
+	gss_channel_bindings_t	 icb;		/* input channel bindings */
+	int			 refcnt;	/* reference count gss AUTHs */
 };
 
 #define	AUTH_PRIVATE(auth)	((struct rpc_gss_data *)auth->ah_private)
 
 static struct timeval AUTH_TIMEOUT = { 25, 0 };
 
+static void
+authgss_auth_get(AUTH *auth)
+{
+	struct rpc_gss_data	*gd = AUTH_PRIVATE(auth);
+
+	mutex_lock(&auth_ref_lock);
+	++gd->refcnt;
+	mutex_unlock(&auth_ref_lock);
+}
+
+static int
+authgss_auth_put(AUTH *auth)
+{
+	int refcnt;
+	struct rpc_gss_data	*gd = AUTH_PRIVATE(auth);
+
+	mutex_lock(&auth_ref_lock);
+	refcnt = --gd->refcnt;
+	mutex_unlock(&auth_ref_lock);
+	return refcnt;
+}
+
 AUTH *
 authgss_create(CLIENT *clnt, gss_name_t name, struct rpc_gss_sec *sec)
 {
@@ -179,6 +207,7 @@ authgss_create(CLIENT *clnt, gss_name_t name, struct rpc_gss_sec *sec)
 			rpc_createerr.cf_stat = RPC_SYSTEMERROR;
 			rpc_createerr.cf_error.re_errno = ENOMEM;
 			free(auth);
+			free(gd);
 			return (NULL);
 		}
 	}
@@ -203,7 +232,7 @@ authgss_create(CLIENT *clnt, gss_name_t name, struct rpc_gss_sec *sec)
 	if (!authgss_refresh(auth, NULL))
 		auth = NULL;
 	else
-		auth_get(auth); /* Reference for caller */
+		authgss_auth_get(auth); /* Reference for caller */
 
 	clnt->cl_auth = save_auth;
 
@@ -448,7 +477,7 @@ _rpc_gss_refresh(AUTH *auth, rpc_gss_options_ret_t *options_ret)
 	for (;;) {
 		/* print the token we just received */
 		if (recv_tokenp != GSS_C_NO_BUFFER) {
-			gss_log_debug("The token we just received (length %d):",
+			gss_log_debug("The token we just received (length %lu):",
 				  recv_tokenp->length);
 			gss_log_hexdump(recv_tokenp->value, recv_tokenp->length, 0);
 		}
@@ -458,8 +487,8 @@ _rpc_gss_refresh(AUTH *auth, rpc_gss_options_ret_t *options_ret)
 						gd->name,
 						gd->sec.mech,
 						gd->sec.req_flags,
-						0,		/* time req */
-						NULL,		/* channel */
+						gd->time_req,
+						gd->icb,
 						recv_tokenp,
 						&actual_mech_type,
 						&send_token,
@@ -481,7 +510,7 @@ _rpc_gss_refresh(AUTH *auth, rpc_gss_options_ret_t *options_ret)
 			memset(&gr, 0, sizeof(gr));
 
 			/* print the token we are about to send */
-			gss_log_debug("The token being sent (length %d):",
+			gss_log_debug("The token being sent (length %lu):",
 				  send_token.length);
 			gss_log_hexdump(send_token.value, send_token.length, 0);
 
@@ -498,6 +527,14 @@ _rpc_gss_refresh(AUTH *auth, rpc_gss_options_ret_t *options_ret)
 			     gr.gr_major != GSS_S_CONTINUE_NEEDED)) {
 				options_ret->major_status = gr.gr_major;
 				options_ret->minor_status = gr.gr_minor;
+				if (call_stat != RPC_SUCCESS) {
+					struct rpc_err err;
+					clnt_geterr(gd->clnt, &err);
+					LIBTIRPC_DEBUG(1, ("authgss_refresh: %s errno: %s",
+						clnt_sperrno(call_stat), strerror(err.re_errno)));
+				} else
+					gss_log_status("authgss_refresh:", 
+						gr.gr_major, gr.gr_minor);
 				return FALSE;
 			}
 
@@ -556,7 +593,7 @@ _rpc_gss_refresh(AUTH *auth, rpc_gss_options_ret_t *options_ret)
 			if (rpc_gss_oid_to_mech(actual_mech_type, &mechanism)) {
 				strncpy(options_ret->actual_mechanism,
 					mechanism,
-					sizeof(options_ret->actual_mechanism));
+					(sizeof(options_ret->actual_mechanism)-1));
 			}
 
 			gd->established = TRUE;
@@ -654,21 +691,24 @@ authgss_destroy_context(AUTH *auth)
 static void
 authgss_destroy(AUTH *auth)
 {
-	struct rpc_gss_data	*gd;
-	OM_uint32		 min_stat;
+	if(authgss_auth_put(auth) == 0)
+	{
+		struct rpc_gss_data	*gd;
+		OM_uint32		 min_stat;
 
-	gss_log_debug("in authgss_destroy()");
+		gss_log_debug("in authgss_destroy()");
 
-	gd = AUTH_PRIVATE(auth);
+		gd = AUTH_PRIVATE(auth);
 
-	authgss_destroy_context(auth);
+		authgss_destroy_context(auth);
 
-	LIBTIRPC_DEBUG(3, ("authgss_destroy: freeing name %p", gd->name));
-	if (gd->name != GSS_C_NO_NAME)
-		gss_release_name(&min_stat, &gd->name);
+		LIBTIRPC_DEBUG(3, ("authgss_destroy: freeing name %p", gd->name));
+		if (gd->name != GSS_C_NO_NAME)
+			gss_release_name(&min_stat, &gd->name);
 
-	free(gd);
-	free(auth);
+		free(gd);
+		free(auth);
+	}
 }
 
 static bool_t
@@ -772,11 +812,6 @@ rpc_gss_seccreate(CLIENT *clnt, char *principal, char *mechanism,
 		return _rpc_gss_seccreate_error(ENOENT);
 	}
 
-	if (req != NULL) {
-		sec.req_flags = req->req_flags;
-		sec.cred = req->my_cred;
-	}
-
 	if (ret == NULL)
 		ret = &options_ret;
 	memset(ret, 0, sizeof(*ret));
@@ -806,6 +841,13 @@ rpc_gss_seccreate(CLIENT *clnt, char *principal, char *mechanism,
 	gd->ctx = GSS_C_NO_CONTEXT;
 	gd->sec = sec;
 
+	if (req) {
+		sec.req_flags = req->req_flags;
+		gd->time_req = req->time_req;
+		sec.cred = req->my_cred;
+		gd->icb = req->input_channel_bindings;
+	}
+
 	gd->gc.gc_v = RPCSEC_GSS_VERSION;
 	gd->gc.gc_proc = RPCSEC_GSS_INIT;
 	gd->gc.gc_svc = sec.svc;
@@ -817,11 +859,10 @@ rpc_gss_seccreate(CLIENT *clnt, char *principal, char *mechanism,
 	clnt->cl_auth = auth;
 
 	if (_rpc_gss_refresh(auth, ret) == FALSE) {
-		authgss_destroy(auth);
 		auth = NULL;
 	} else {
 		rpc_gss_clear_error();
-		auth_get(auth);
+		authgss_auth_get(auth);
 	}
 
 	clnt->cl_auth = save_auth;
